#
# Used during the release process to make sure that we release based on a
# tag that has the same version as the current twisted.__version.
#
# Designed to be conditionally called inside GitHub Actions release job.
# Tags should use PEP440 version scheme.
#
# To be called as: admin/check_tag_version_match.py refs/tags/twisted-20.3.0
#
import sys

import pep517.meta

TAG_PREFIX = "refs/tags/twisted-"

if len(sys.argv) < 2:
    print("No tag check requested.")
    sys.exit(0)

branch_version = pep517.meta.load(".").version
run_version = sys.argv[1]

if not run_version.startswith(TAG_PREFIX):
    print(f"Not a twisted release tag name '{run_version}.")
    sys.exit(1)

run_version = run_version[len(TAG_PREFIX) :]

if run_version != branch_version:
    print(f"Branch is at '{branch_version}' while tag is '{run_version}'")
    exit(1)

print(f"All good. Branch and tag versions match for '{branch_version}'.")
sys.exit(0)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.
from hatchling.metadata.plugin.interface import MetadataHookInterface


class CustomMetadataHook(MetadataHookInterface):
    def update(self, metadata):
        optional_dependencies = self.config["optional-dependencies"]
        for feature, dependencies in list(optional_dependencies.items()):
            if "-" not in feature:
                continue

            # We define all the legacy dependency targets using the
            # underscore for backward compatibility.
            #
            # See: https://github.com/twisted/twisted/pull/11656#issuecomment-1282855123
            legacy_feature = feature.replace("-", "_")
            optional_dependencies[legacy_feature] = dependencies

        metadata["optional-dependencies"] = optional_dependencies



import importlib.metadata
import os
import sys

print("::group::Python info")
print(sys.prefix)
print(sys.exec_prefix)
print(sys.executable)
print(sys.version)
print(sys.platform)
print("::endgroup::")

if sys.version_info >= (3, 10):
    _entry_points = importlib.metadata.entry_points
else:

    def _entry_points(*, group: str, name: str):
        for ep in importlib.metadata.entry_points().get(group, []):
            if ep.name == name:
                yield ep


if os.environ.get("CI", "").lower() == "true":
    # On CI show the exact deps found at runtime.
    # Skipped on local dev to speed up the test execution.
    print("::group::Deps info")
    for ep in _entry_points(group="console_scripts", name="pip"):
        ep.load()(["freeze", "--all"])
        break
    print("::endgroup::")



import pwd

from twisted.internet import defer, protocol, reactor, utils

# Another back-end


@implementer(IFingerService)
class LocalFingerService(service.Service):
    def getUser(self, user):
        # need a local finger daemon running for this to work
        return utils.getProcessOutput("finger", [user])

    def getUsers(self):
        return defer.succeed([])


f = LocalFingerService()



from twisted.internet import endpoints, protocol, reactor
from twisted.protocols import basic


class FingerProtocol(basic.LineReceiver):
    def lineReceived(self, user):
        self.transport.write(b"No such user\r\n")
        self.transport.loseConnection()


class FingerFactory(protocol.ServerFactory):
    protocol = FingerProtocol


fingerEndpoint = endpoints.serverFromString(reactor, "tcp:1079")
fingerEndpoint.listen(FingerFactory())
reactor.run()



from twisted.internet import endpoints, protocol, reactor


class FingerProtocol(protocol.Protocol):
    pass


class FingerFactory(protocol.ServerFactory):
    protocol = FingerProtocol


fingerEndpoint = endpoints.serverFromString(reactor, "tcp:1079")
fingerEndpoint.listen(FingerFactory())
reactor.run()



from http.cookiejar import CookieJar

from twisted.internet import reactor
from twisted.python import log
from twisted.web.client import Agent, CookieAgent


def displayCookies(response, cookieJar):
    print("Received response")
    print(response)
    print("Cookies:", len(cookieJar))
    for cookie in cookieJar:
        print(cookie)


def main():
    cookieJar = CookieJar()
    agent = CookieAgent(Agent(reactor), cookieJar)

    d = agent.request(b"GET", b"http://httpbin.org/cookies/set?some=data")
    d.addCallback(displayCookies, cookieJar)
    d.addErrback(log.err)
    d.addCallback(lambda ignored: reactor.stop())
    reactor.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any

from twisted.internet.defer import Failure
from twisted.internet.task import LoopingCall, deferLater, react
from twisted.web.websocket import WebSocketClientEndpoint, WebSocketTransport


class WebSocketClientDemo:
    @classmethod
    def buildProtocol(cls, uri: str) -> WebSocketClientDemo:
        return cls()

    def textMessageReceived(self, data: str) -> None:
        print(f"received text: {data!r}")

    def negotiationStarted(self, transport: WebSocketTransport) -> None:
        self.transport = transport

    def negotiationFinished(self) -> None:
        self.transport.sendTextMessage("hello, world!")

    def bytesMessageReceived(self, data: bytes) -> None:
        ...

    def connectionLost(self, reason: Failure) -> None:
        ...

    def pongReceived(self, payload: bytes) -> None:
        ...


async def main(reactor: Any) -> None:
    endpoint = WebSocketClientEndpoint.new(
        reactor, "ws://localhost:8080/websocket-server.rpy"
    )
    print("connecting...")
    await endpoint.connect(WebSocketClientDemo)
    print("connected!")
    await deferLater(reactor, 10)


react(main)



from twisted.internet.defer import Deferred, ensureDeferred, succeed


def test_deferred_await(benchmark):
    """Measure the speed of awaiting and of defer.succeed()."""

    async def _run():
        for x in range(1000):
            await succeed(x)

    def go():
        ensureDeferred(_run())

    benchmark(go)


def test_deferred_await_unfired(benchmark):
    """Measure the speed of awaiting unfired Deferreds."""

    async def _run(deferreds):
        results = []
        for d in deferreds:
            results.append(await d)
        return sum(results)

    def go():
        deferreds = [Deferred() for _ in range(20)]
        result = ensureDeferred(_run(deferreds))
        for i in range(20):
            deferreds[i].callback(i)
        results = []
        result.addCallback(results.append)
        assert results[0] == sum(range(20))

    benchmark(go)


def f(x: int, a: int) -> int:
    return x + a


def test_deferred_callback_chain_fires_at_start(benchmark):
    """
    Measure speed of successful callbacks, where the Deferred fires before the
    callbacks are added.
    """

    def go():
        d = Deferred()
        d.callback(2)
        d.addCallback(f, 1)
        d.addCallback(f, 2)
        d.addCallback(f, 3)
        d.addCallback(f, 4)

    benchmark(go)


def test_deferred_callback_chain_fires_at_end(benchmark):
    """
    Measure speed of successful callbacks, where the Deferred fires after the
    callbacks are added.
    """

    def go():
        d = Deferred()
        d.addCallback(f, 1)
        d.addCallback(f, 2)
        d.addCallback(f, 3)
        d.addCallback(f, 4)
        d.callback(2)

    benchmark(go)


def test_deferred_errback_chain(benchmark):
    """
    Measure speed of error handling in callbacks.
    callbacks are added.
    """

    def go():
        d = succeed("result")

        def cbRaiseErr(_):
            raise Exception("boom!")

        d.addCallback(cbRaiseErr)

        def ebHandleErr(failure):
            failure.trap(Exception)
            raise Exception("lesser boom!")

        d.addErrback(ebHandleErr)

        def swallowErr(_):
            return None

        d.addBoth(swallowErr)

    benchmark(go)


def test_deferred_chained_already_fired(benchmark):
    """
    Measure speed of chained Deferred, where the chained Deferred fires before
    the callback returning it is added.
    """

    def go():
        d = Deferred()
        d2 = Deferred()
        d3 = Deferred()
        d2.callback(123)
        d3.callback(456)
        d.addCallback(lambda _: d2)
        d.addCallback(lambda _: d3)
        d.callback(123)
        result = []
        d.addCallback(result.append)
        assert result == [456]

    benchmark(go)


def test_deferred_chained_not_fired(benchmark):
    """
    Measure speed of chained Deferred, where the chained Deferred fires after
    the callback returning it is added.
    """

    def go():
        d = Deferred()
        d2 = Deferred()
        # d3 has its own chained result:
        d3 = Deferred()
        d4 = Deferred()
        d3.addCallback(lambda _: d4)
        d3.callback(123)
        d.addCallback(lambda _: d2)
        d.addCallback(lambda _: d3)
        d2.callback(123)
        d4.callback(57)
        d.callback(7)
        result = []
        d.addCallback(result.append)
        assert result == [57]

    benchmark(go)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Send a HTTP request to Docker over a Unix socket.

Will probably need to be run as root.

Usage:
    $ sudo python endpointconstructor.py [<docker API path>]
"""


from sys import argv

from zope.interface import implementer

from twisted.internet.endpoints import UNIXClientEndpoint
from twisted.internet.task import react
from twisted.web.client import Agent, readBody
from twisted.web.iweb import IAgentEndpointFactory


@implementer(IAgentEndpointFactory)
class DockerEndpointFactory:
    """
    Connect to Docker's Unix socket.
    """

    def __init__(self, reactor):
        self.reactor = reactor

    def endpointForURI(self, uri):
        return UNIXClientEndpoint(self.reactor, b"/var/run/docker.sock")


def main(reactor, path=b"/containers/json?all=1"):
    agent = Agent.usingEndpointFactory(reactor, DockerEndpointFactory(reactor))
    d = agent.request(b"GET", b"unix://localhost" + path)
    d.addCallback(readBody)
    d.addCallback(print)
    return d


react(main, argv[1:])



from calculus.client_2 import RemoteCalculationClient
from calculus.remote_1 import RemoteCalculationFactory

from twisted.internet import protocol, reactor
from twisted.trial import unittest


class RemoteRunCalculationTestCase(unittest.TestCase):
    def setUp(self):
        factory = RemoteCalculationFactory()
        self.port = reactor.listenTCP(0, factory, interface="127.0.0.1")
        self.client = None

    def tearDown(self):
        if self.client is not None:
            self.client.transport.loseConnection()
        return self.port.stopListening()

    def _test(self, op, a, b, expected):
        creator = protocol.ClientCreator(reactor, RemoteCalculationClient)

        def cb(client):
            self.client = client
            return getattr(self.client, op)(a, b).addCallback(
                self.assertEqual, expected
            )

        return creator.connectTCP("127.0.0.1", self.port.getHost().port).addCallback(cb)

    def test_add(self):
        return self._test("add", 5, 9, 14)

    def test_subtract(self):
        return self._test("subtract", 47, 13, 34)

    def test_multiply(self):
        return self._test("multiply", 7, 3, 21)

    def test_divide(self):
        return self._test("divide", 84, 10, 8)



# -*- test-case-name: calculus.test.test_client_1 -*-

from twisted.internet import defer
from twisted.protocols import basic


class RemoteCalculationClient(basic.LineReceiver):
    def __init__(self):
        self.results = []

    def lineReceived(self, line):
        d = self.results.pop(0)
        d.callback(int(line))

    def _sendOperation(self, op, a, b):
        d = defer.Deferred()
        self.results.append(d)
        line = f"{op} {a} {b}".encode()
        self.sendLine(line)
        return d

    def add(self, a, b):
        return self._sendOperation("add", a, b)

    def subtract(self, a, b):
        return self._sendOperation("subtract", a, b)

    def multiply(self, a, b):
        return self._sendOperation("multiply", a, b)

    def divide(self, a, b):
        return self._sendOperation("divide", a, b)



#!/usr/bin/env python
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.
"""
tls_alpn_server
~~~~~~~~~~~~~~~~~~~

This test script demonstrates the usage of the acceptableProtocols API as a
server peer.

It performs next protocol negotiation using ALPN.

It will print what protocol was negotiated for each connection that is made to
it.

To exit the server, use CTRL+C on the command-line.

Before using this, you should generate a new RSA private key and an associated
X.509 certificate and place it in the working directory as `server-key.pem` and
`server-cert.pem`.

You can generate a self signed certificate using OpenSSL::

    openssl req -new -newkey rsa:2048 -days 3 -nodes -x509 \
        -keyout server-key.pem -out server-cert.pem

To test this, use OpenSSL's s_client command with the -alpn argument.

For example::

    openssl s_client -connect localhost:8080 -alpn h2,http/1.1

Alternatively, use the tls_alpn_client.py script found in the examples
directory.
"""

from OpenSSL import crypto

from twisted.internet import reactor, ssl
from twisted.internet.endpoints import SSL4ServerEndpoint
from twisted.internet.protocol import Factory, Protocol
from twisted.python.filepath import FilePath

# The list of protocols we'd be prepared to speak after the TLS negotiation is
# complete.
# The order of the protocols here is an order of preference. This indicates
# that we'd prefer to use HTTP/2 if possible (where HTTP/2 is using the token
# 'h2'), but would also accept HTTP/1.1.
# The bytes here are sent literally on the wire, and so there is no room for
# ambiguity about text encodings.
# Try changing this list by adding, removing, and reordering protocols to see
# how it affects the result.
ACCEPTABLE_PROTOCOLS = [b"h2", b"http/1.1"]

# The port that the server will listen on.
LISTEN_PORT = 8080


class ShowALPN(Protocol):
    """
    This protocol accepts incoming connections and waits for data. When
    received, it prints what the negotiated protocol is, echoes the data back,
    and then terminates the connection.
    """

    def connectionMade(self):
        self.complete = False
        print("Connection made")

    def dataReceived(self, data):
        print(self.transport.negotiatedProtocol)
        self.transport.write(data)
        self.complete = True
        self.transport.loseConnection()

    def connectionLost(self, reason):
        # If we haven't received any data, an error occurred. Otherwise,
        # we lost the connection on purpose.
        if self.complete:
            print("Connection closed cleanly")
        else:
            print(f"Connection lost due to error {reason}")


class ResponderFactory(Factory):
    def buildProtocol(self, addr):
        return ShowALPN()


privateKeyData = FilePath("server-key.pem").getContent()
privateKey = crypto.load_privatekey(crypto.FILETYPE_PEM, privateKeyData)
certData = FilePath("server-cert.pem").getContent()
certificate = crypto.load_certificate(crypto.FILETYPE_PEM, certData)

options = ssl.CertificateOptions(
    privateKey=privateKey,
    certificate=certificate,
    acceptableProtocols=ACCEPTABLE_PROTOCOLS,
)
endpoint = SSL4ServerEndpoint(reactor, LISTEN_PORT, options)
endpoint.listen(ResponderFactory())
reactor.run()



#!/usr/bin/env python

"""
Twisted moved the C{twisted} hierarchy to the C{src} hierarchy, but C{git}
doesn't know how to track moves of directories, only files.  Therefore any
files added in branches after this move will be added into ./twisted/ and need
to be moved over into.
"""

import os

from twisted.python.filepath import FilePath

here = FilePath(__file__).parent().parent()
fromPath = here.child("twisted")
toPath = here.child("src")

for fn in fromPath.walk():
    if fn.isfile():
        os.system("git mv {it} src/{it}".format(it="/".join(fn.segmentsFrom(here))))

os.system("git clean -fd")



#!/usr/bin/env python

# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


from twisted.internet import reactor
from twisted.spread import pb


class One(pb.Root):
    def remote_takeTwo(self, two):
        print("received a Two called", two)
        print("telling it to print(12)")
        two.callRemote("print", 12)


reactor.listenTCP(8800, pb.PBServerFactory(One()))
reactor.run()



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
This example demonstrates how to use downloadPage.

Usage:
    $ python dlpage.py <url>

Don't forget the http:// when you type a URL!
"""

import sys

from twisted.internet import reactor
from twisted.python.util import println
from twisted.web.client import downloadPage

# The function downloads a page and saves it to a file, in this case, it saves
# the page to "foo".
url = sys.argv[1].encode("ascii")
downloadPage(url, "foo").addCallbacks(
    lambda value: reactor.stop(),
    lambda error: (println("an error occurred", error), reactor.stop()),
)
reactor.run()



"""Tests for scheduled events."""

import pytest

from twisted.internet.base import ReactorBase


class ReactorWithRiggedTime(ReactorBase):
    """
    Allows controlling the passed time while benchmarking C{ReactorBase}.
    """

    _currentTime = 0.0

    def seconds(self) -> float:
        """
        Override the regular time function in C{ReactorBase}.
        """
        return self._currentTime

    def advance(self, seconds: float) -> None:
        """
        Advance time by the given numbe of seconds.
        """
        self._currentTime += seconds

    def installWaker(self) -> None:
        """
        Subclasses of C{ReactorBase} are required to implement this.
        """
        pass


@pytest.mark.parametrize("reversed_times", [False, True])
def test_spaced_out_events(benchmark, reversed_times):
    """
    Add spaced out events, then run them all.
    """
    timestamps = list(range(100))
    if reversed_times:
        timestamps.reverse()

    def go():
        reactor = ReactorWithRiggedTime()
        for time in timestamps:
            reactor.callLater(time, lambda: None)
        for _ in range(len(timestamps)):
            reactor.advance(reactor.timeout() or 0)
            reactor.runUntilCurrent()
        assert not reactor._pendingTimedCalls
        assert not reactor._newTimedCalls

    benchmark(go)



#!/usr/bin/env python

"""
Twisted moved the C{twisted} hierarchy to the C{src} hierarchy, but C{git}
doesn't know how to track moves of directories, only files.  Therefore any
files added in branches after this move will be added into ./twisted/ and need
to be moved over into
"""

import os

from twisted.python.filepath import FilePath

here = FilePath(__file__).parent().parent()
twistedPath = here.child("src").child("twisted")


def mv(fromPath):
    for fn in fromPath.walk():
        if fn.isfile():
            os.system(
                "git mv {fr} {to}".format(
                    fr=fn.path,
                    to=fn.parent()
                    .parent()
                    .child("newsfragments")
                    .child(fn.basename())
                    .path,
                )
            )


if twistedPath.child("topfiles").exists():
    mv(twistedPath.child("topfiles"))

for child in twistedPath.listdir():
    path = twistedPath.child(child).child("topfiles")
    if path.exists():
        mv(path)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Benchmarks for C{HTTP11ClientProtocol}.
"""

from twisted.internet.testing import StringTransport
from twisted.web._newclient import HTTP11ClientProtocol, Request
from twisted.web.http_headers import Headers

RESPONSE = """HTTP/1.1 200 OK
Host: blah
Foo: bar
Gaz: baz
Content-length: 3

abc""".replace(
    "\n", "\r\n"
).encode(
    "utf-8"
)


def test_http_client_small_response(benchmark):
    """Measure the time to run a simple HTTP 1.1 client request."""

    def go():
        protocol = HTTP11ClientProtocol()
        protocol.makeConnection(StringTransport())
        request = Request(
            b"GET", b"/foo/bar", Headers({b"Host": [b"example.com"]}), None
        )
        response = protocol.request(request)
        protocol.dataReceived(RESPONSE)
        result = []
        response.addCallback(result.append)
        assert result

    benchmark(go)



"""
Benchmarks for LogPublisher event dispatching.
"""

from twisted.logger import LogEvent, LogLevel, LogPublisher


class DummyObserver:
    """
    An observer that just records the last event it received.
    """

    def __init__(self):
        self.last_event = None

    def __call__(self, event: LogEvent):
        self.last_event = event


def test_log_publisher_call_dispatch(benchmark):
    """
    Dispatch a single event to a single publisher that has multiple observers.
    """
    num_observers = 2000

    observers: list[DummyObserver] = [DummyObserver() for _ in range(num_observers)]
    publisher = LogPublisher(*observers)

    event = {
        "log_level": LogLevel.info,
        "log_namespace": "benchmark",
        "log_format": "This is a benchmark event.",
    }

    def go():
        publisher(event)

    benchmark(go)


def test_log_publisher_add_remove(benchmark):
    """
    Benchmark the cost of adding and removing observers repeatedly.
    """
    num_observers = 2000

    publisher = LogPublisher()
    observers = [DummyObserver() for _ in range(num_observers)]

    def go():
        for obs in observers:
            publisher.addObserver(obs)
        for obs in observers:
            publisher.removeObserver(obs)

    benchmark(go)



from twisted.python.filepath import FilePath
from twisted.web.template import Element, XMLFile, renderer


class ExampleElement(Element):
    loader = XMLFile(FilePath("template-1.xml"))

    @renderer
    def header(self, request, tag):
        return tag("Header.")

    @renderer
    def footer(self, request, tag):
        return tag("Footer.")



#!/usr/bin/env python
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

import sys

import echoserv

from twisted.internet import defer, protocol, ssl, task
from twisted.internet.endpoints import TCP6ServerEndpoint, wrapServerTLS
from twisted.python import log
from twisted.python.modules import getModule


def main(reactor):
    log.startLogging(sys.stdout)
    certData = getModule(__name__).filePath.sibling("server.pem").getContent()
    authData = getModule(__name__).filePath.sibling("server.pem").getContent()
    certificate = ssl.PrivateCertificate.loadPEM(certData)
    authority = ssl.Certificate.loadPEM(certData)
    factory = protocol.Factory.forProtocol(echoserv.Echo)
    tcpEndpoint = TCP6ServerEndpoint(reactor, 8000)
    endpoint = wrapServerTLS(certificate.options(authority), tcpEndpoint, reactor)
    endpoint.listen(factory)
    return defer.Deferred()


if __name__ == "__main__":
    import ssl_clientauth_server

    task.react(ssl_clientauth_server.main)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Benchmarks for the C{twisted.web} server.
"""

from twisted.internet.testing import StringTransport
from twisted.web import resource, server


class Data(resource.Resource):
    """
    This is a static, in-memory resource.
    """

    # Being a leaf, we will not have getChild called on this resource.
    isLeaf = True

    def __init__(self, data, type):
        resource.Resource.__init__(self)
        self.data = data
        self.data_len = b"%d" % (len(self.data),)
        self.type = type

    def render_GET(self, request):
        request.setHeader(b"content-type", self.type)
        request.setHeader(b"content-length", self.data_len)
        return self.data


class ComplexData(Data):
    """
    Interact more with the request.
    """

    def render_GET(self, request):
        request.setLastModified(123)
        request.setETag(b"xykjlk")
        _ = request.getRequestHostname()
        request.setHost(b"example.com", 80)
        return Data.render_GET(self, request)


def http11_server_empty_request(resource, benchmark):
    """Benchmark of handling an bodyless HTTP/1.1 request."""
    factory = server.Site(resource)

    def go():
        transport = StringTransport()
        protocol = factory.buildProtocol(None)
        protocol.makeConnection(transport)
        protocol.dataReceived(
            b"""\
GET / HTTP/1.1
Host: example.com
User-Agent: XXX
Time: XXXX
Content-Length: 0

""".replace(
                b"\n", b"\r\n"
            )
        )
        assert b"200 OK" in transport.io.getvalue()

    benchmark(go)


def test_http11_server_empty_request(benchmark):
    """Benchmark just returning some data."""
    data = Data(b"This is a result hello hello" * 4, b"text/plain")
    http11_server_empty_request(data, benchmark)


def test_bit_more_complex_response(benchmark):
    """Benchmark that also involves calling more request methods."""
    data = ComplexData(b"This is a result hello hello" * 4, b"text/plain")
    http11_server_empty_request(data, benchmark)


def test_http11_server_many_headers(benchmark):
    """Benchmark handling of an HTTP/1.1 request with many headers."""
    request = b"\r\n".join(
        [
            b"GET / HTTP/1.1",
            b"Host: example.com",
        ]
        + [f"X-{name}: {name}".encode() for name in ("foo", "bar", "Baz", "biff")]
        + [f"x-tab:   {name}\t{name}".encode() for name in ("x", "y", "z", "q")]
        + [
            f"Cookie:   {c}={c * 26}".encode()
            for c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ]
        + [b"", b""]
    )
    data = Data(b"...\n", b"text/plain")
    factory = server.Site(data)

    def go():
        transport = StringTransport()
        protocol = factory.buildProtocol(None)
        protocol.makeConnection(transport)
        protocol.dataReceived(request)
        assert b"200 OK" in transport.io.getvalue()

    benchmark(go)


def test_http11_server_chunked_request(benchmark):
    """
    Benchmark receipt of a largeish chunked request.
    """
    request = (
        b"""\
GET / HTTP/1.1
Host: example.com
user-agent: XXX
Transfer-encoding: chunked

""".replace(
            b"\n", b"\r\n"
        )
        + b"d\r\nHello, world!\r\n" * 100
        + b"0\r\n\r\n"
    )
    data = Data(b"Goodbye!\n", b"text/plain")
    factory = server.Site(data)

    def go():
        transport = StringTransport()
        protocol = factory.buildProtocol(None)
        protocol.makeConnection(transport)
        protocol.dataReceived(request)
        assert transport.io.getvalue().startswith(b"HTTP/1.1 200 ")

    benchmark(go)


class Chunker(resource.Resource):
    """
    Static data that is written out in chunks.
    """

    isLeaf = True  # no getChild

    def __init__(self, chunks, type):
        resource.Resource.__init__(self)
        self.chunks = chunks
        self.type = type

    def render_GET(self, request):
        request.setHeader(b"Content-Type", self.type)
        for chunk in self.chunks:
            request.write(chunk)
        request.finish()
        return server.NOT_DONE_YET


class BitbucketTransport(StringTransport):
    """
    Throw away any data written, in Web Scale (TM) fashion.
    """

    def write(self, data):
        """
        Carefully ensure that nothing is done with the data.
        """

    def writeSequence(self, iovec):
        """
        Lovingly misplace the data.
        """
        # This override method is necessary because the superclass
        # method is implemented as self.write(b"".join(iovec)),
        # which does too much work!


def test_http11_server_chunked_response(benchmark):
    """
    Benchmark generation of a largeish chunked response.
    """
    data = Chunker(
        [
            bytes([c]) * 1024
            for c in b"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ],
        b"application/octet-stream",
    )
    factory = server.Site(data)

    def go(*, transportFactory=BitbucketTransport):
        transport = transportFactory()
        protocol = factory.buildProtocol(None)
        protocol.makeConnection(transport)
        protocol.dataReceived(
            b"""\
GET / HTTP/1.1
host: example.com
accept: *

""".replace(
                b"\n", b"\r\n"
            )
        )
        return transport

    # Sanity-check that the benchmark logic produces a chunked response.
    transport = go(transportFactory=StringTransport)
    response = transport.io.getvalue()
    assert response.startswith(b"HTTP/1.1 200 ")
    assert b"Transfer-Encoding: chunked" in response[:1024]

    # Now benchmark, discarding all the HTTP response bytes.
    benchmark(go)



"""
Benchmarks for line parsing protocols.
"""

import pytest

from twisted.internet.protocol import Protocol
from twisted.internet.testing import StringTransport
from twisted.protocols.basic import LineOnlyReceiver, LineReceiver


def deliverData(protocol: Protocol, data: bytes, chunkSize: int) -> None:
    i = 0
    while i < len(data):
        protocol.dataReceived(data[i : i + chunkSize])
        i += chunkSize


@pytest.mark.parametrize("chunkSize", [16, 64, 256, 1024])
def test_lineReceiver(benchmark, chunkSize):
    """
    Parses lines, but can in theory also parse raw data.
    """

    class MyLineReceiver(LineReceiver):
        def lineReceived(self, _):
            pass

    protocol = MyLineReceiver()
    protocol.makeConnection(StringTransport())
    data = ((b"abcde" * 10) + b"\r\n") * 1000

    benchmark(lambda: deliverData(protocol, data, chunkSize))


@pytest.mark.parametrize("chunkSize", [16, 64, 256, 1024])
def test_lineOnlyReceiver(benchmark, chunkSize):
    """
    Parses only lines.
    """

    class MyLineReceiver(LineOnlyReceiver):
        def lineReceived(self, line):
            pass

    protocol = MyLineReceiver()
    protocol.makeConnection(StringTransport())
    data = ((b"abcde" * 10) + b"\r\n") * 1000

    benchmark(lambda: deliverData(protocol, data, chunkSize))



from sys import argv

from twisted.internet.defer import Deferred
from twisted.internet.endpoints import HostnameEndpoint, wrapClientTLS
from twisted.internet.interfaces import IReactorTCP, ITCPTransport
from twisted.internet.protocol import Factory, Protocol
from twisted.internet.ssl import optionsForClientTLS
from twisted.internet.task import react
from twisted.python.failure import Failure


async def main(reactor: IReactorTCP, hostname: str = "example.com") -> None:
    class ExampleHTTP(Protocol):
        def makeConnection(self, transport: ITCPTransport) -> None:
            transport.write(f"GET / HTTP/1.1\r\nHost: {hostname}\r\n\r\n".encode())

        def dataReceived(self, data: bytes) -> None:
            print(f"data: {data!r}")

    tcpEndpoint = HostnameEndpoint(reactor, hostname, 443)
    tlsEndpoint = wrapClientTLS(optionsForClientTLS(hostname), tcpEndpoint)
    await tlsEndpoint.connect(Factory.forProtocol(ExampleHTTP))
    await Deferred()


if __name__ == "__main__":
    react(main, argv[1:])



import time

from twisted.protocols import basic
from twisted.python.compat import range


class CollectingLineReceiver(basic.LineReceiver):
    def __init__(self):
        self.lines = []
        self.lineReceived = self.lines.append


def deliver(proto, chunks):
    return [proto.dataReceived(chunk) for chunk in chunks]


def benchmark(chunkSize, lineLength, numLines):
    bytes = (b"x" * lineLength + b"\r\n") * numLines
    chunkCount = len(bytes) // chunkSize + 1
    chunks = []
    for n in range(chunkCount):
        chunks.append(bytes[n * chunkSize : (n + 1) * chunkSize])
    assert b"".join(chunks) == bytes, (chunks, bytes)
    p = CollectingLineReceiver()

    before = time.clock()
    deliver(p, chunks)
    after = time.clock()

    assert bytes.splitlines() == p.lines, (bytes.splitlines(), p.lines)

    print("chunkSize:", chunkSize, end=" ")
    print("lineLength:", lineLength, end=" ")
    print("numLines:", numLines, end=" ")
    print("CPU Time: ", after - before)


def main():
    for numLines in 100, 1000:
        for lineLength in (10, 100, 1000):
            for chunkSize in (1, 500, 5000):
                benchmark(chunkSize, lineLength, numLines)

    for numLines in 10000, 50000:
        for lineLength in (1000, 2000):
            for chunkSize in (51, 500, 5000):
                benchmark(chunkSize, lineLength, numLines)


if __name__ == "__main__":
    main()



from twisted.internet import endpoints
from twisted.web import server, xmlrpc


class EchoHandler:
    def echo(self, x):
        """
        Return all passed args
        """
        return x


class AddHandler:
    def add(self, a, b):
        """
        Return sum of arguments.
        """
        return a + b


class Example(xmlrpc.XMLRPC):
    """
    An example of using you own policy to fetch the handler
    """

    def __init__(self):
        xmlrpc.XMLRPC.__init__(self)
        self._addHandler = AddHandler()
        self._echoHandler = EchoHandler()

        # We keep a dict of all relevant
        # procedure names and callable.
        self._procedureToCallable = {
            "add": self._addHandler.add,
            "echo": self._echoHandler.echo,
        }

    def lookupProcedure(self, procedurePath):
        try:
            return self._procedureToCallable[procedurePath]
        except KeyError as e:
            raise xmlrpc.NoSuchFunction(
                self.NOT_FOUND, "procedure %s not found" % procedurePath
            )

    def listProcedures(self):
        """
        Since we override lookupProcedure, its suggested to override
        listProcedures too.
        """
        return ["add", "echo"]


if __name__ == "__main__":
    from twisted.internet import reactor

    r = Example()
    endpoint = endpoints.TCP4ServerEndpoint(reactor, 7080)
    endpoint.listen(server.Site(r))
    reactor.run()



from calculus.client_1 import RemoteCalculationClient

from twisted.test import proto_helpers
from twisted.trial import unittest


class ClientCalculationTestCase(unittest.TestCase):
    def setUp(self):
        self.tr = proto_helpers.StringTransport()
        self.proto = RemoteCalculationClient()
        self.proto.makeConnection(self.tr)

    def _test(self, operation, a, b, expected):
        d = getattr(self.proto, operation)(a, b)
        self.assertEqual(self.tr.value(), f"{operation} {a} {b}\r\n".encode())
        self.tr.clear()
        d.addCallback(self.assertEqual, expected)
        self.proto.dataReceived(
            "{}\r\n".format(
                expected,
            ).encode("utf-8")
        )
        return d

    def test_add(self):
        return self._test("add", 7, 6, 13)

    def test_subtract(self):
        return self._test("subtract", 82, 78, 4)

    def test_multiply(self):
        return self._test("multiply", 2, 8, 16)

    def test_divide(self):
        return self._test("divide", 14, 3, 4)



from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.protocol import Factory, Protocol, ServerFactory
from twisted.internet.testing import _benchmarkWithReactor as benchmarkWithReactor
from twisted.protocols.wire import Echo


class Server(Echo):
    def connectionMade(self):
        self.transport.setTcpNoDelay(True)


class Counter(Protocol):
    count = 0
    sendAmount = 0
    finished: None | Deferred = None

    def connectionMade(self):
        self.transport.setTcpNoDelay(True)

    def dataReceived(self, b):
        self.count += len(b)
        if self.count == self.sendAmount:
            self.transport.loseConnection()

    def connectionLost(self, reason):
        assert self.finished is not None
        self.finished.callback(self.count)


class Client:
    def __init__(self, reactor, server):
        self._reactor = reactor
        self._server = server
        self._sent = 0

    def run(self, sendAmount, chunkSize):
        self._sendAmount = sendAmount
        self._bytes = b"x" * chunkSize
        # Set up a connection
        factory = Factory()
        factory.protocol = Counter
        d = self._server.connect(factory)
        d.addCallback(self._connected)
        return d

    def _connected(self, client):
        self._client = client
        self._client.finished = Deferred()
        self._client.sendAmount = self._sendAmount
        client.transport.registerProducer(self, False)
        return self._client.finished

    def resumeProducing(self):
        self._client.transport.write(self._bytes)
        self._sent += len(self._bytes)
        if self._sent == self._sendAmount:
            self._client.transport.unregisterProducer()


@benchmarkWithReactor
async def test_tcp_throughput():
    chunkSize = 16384
    sendAmount = 1024 * 1024

    server = ServerFactory()
    server.protocol = Server
    port = reactor.listenTCP(0, server)
    client = Client(
        reactor, TCP4ClientEndpoint(reactor, "127.0.0.1", port.getHost().port)
    )
    result = await client.run(sendAmount, chunkSize)
    await port.stopListening()
    assert result == sendAmount



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Benchmark for SSH connection setup between a Conch client and server using RSA
keys.
"""
from twisted.conch.ssh.factory import SSHFactory
from twisted.conch.ssh.keys import Key
from twisted.conch.ssh.transport import SSHClientTransport
from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.internet.endpoints import (
    TCP4ClientEndpoint,
    TCP4ServerEndpoint,
    connectProtocol,
)
from twisted.internet.testing import _benchmarkWithReactor as benchmarkWithReactor

PUBLIC_KEY = (
    b"ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAGEArzJx8OYOnJmzf4tfBEvLi8DVPrJ3/c9k2I/Az6"
    b"4fxjHf9imyRJbixtQhlH9lfNjUIx+4LmrJH5QNRsFporcHDKOTwTTYLh5KmRpslkYHRivcJSkb"
    b"h/C+BR3utDS555mV"
)

PRIVATE_KEY = b"""-----BEGIN RSA PRIVATE KEY-----
MIIByAIBAAJhAK8ycfDmDpyZs3+LXwRLy4vA1T6yd/3PZNiPwM+uH8Yx3/YpskSW
4sbUIZR/ZXzY1CMfuC5qyR+UDUbBaaK3Bwyjk8E02C4eSpkabJZGB0Yr3CUpG4fw
vgUd7rQ0ueeZlQIBIwJgbh+1VZfr7WftK5lu7MHtqE1S1vPWZQYE3+VUn8yJADyb
Z4fsZaCrzW9lkIqXkE3GIY+ojdhZhkO1gbG0118sIgphwSWKRxK0mvh6ERxKqIt1
xJEJO74EykXZV4oNJ8sjAjEA3J9r2ZghVhGN6V8DnQrTk24Td0E8hU8AcP0FVP+8
PQm/g/aXf2QQkQT+omdHVEJrAjEAy0pL0EBH6EVS98evDCBtQw22OZT52qXlAwZ2
gyTriKFVoqjeEjt3SZKKqXHSApP/AjBLpF99zcJJZRq2abgYlf9lv1chkrWqDHUu
DZttmYJeEfiFBBavVYIF1dOlZT0G8jMCMBc7sOSZodFnAiryP+Qg9otSBjJ3bQML
pSTqy7c3a2AScC/YyOwkDaICHnnD3XyjMwIxALRzl0tQEKMXs6hH8ToUdlLROCrP
EhQ0wahUTCk1gKA4uPD6TMTChavbh4K63OvbKg==
-----END RSA PRIVATE KEY-----"""


class BenchmarkSSHClientTransport(SSHClientTransport):
    """
    The client used in the tests.
    """

    def verifyHostKey(self, hostKey, fingerprint):
        """
        For this test, we don't validate the server identity.
        """
        return succeed(True)

    def connectionSecure(self):
        """
        As soon as the SSH handshake is done, we disconnect,
        without requesting for any SSH service.
        """
        self.transport.loseConnection()

    def connectionLost(self, reason):
        self.factory.clientDisconnected.callback(None)


class BenchmarkSSHServerFactory(SSHFactory):
    """
    A simple SSH server factory with static RSA keys.
    """

    publicKeys = {b"ssh-rsa": Key.fromString(data=PUBLIC_KEY)}
    privateKeys = {b"ssh-rsa": Key.fromString(data=PRIVATE_KEY)}

    def __init__(self):
        # Called by the client once disconnected.
        self.clientDisconnected = Deferred()


@benchmarkWithReactor
async def test_connect_and_disconnect():
    """
    This is the test for key exchange for both client and server.
    Once KEX is done the client disconnects.
    """
    serverFactory = BenchmarkSSHServerFactory()
    serverEndpoint = TCP4ServerEndpoint(reactor, 0)
    serverPort = await serverEndpoint.listen(serverFactory)

    clientProtocol = BenchmarkSSHClientTransport()
    clientProtocol.factory = serverFactory

    clientEndpoint = TCP4ClientEndpoint(
        reactor,
        "127.0.0.1",
        serverPort.getHost().port,
        timeout=5,
    )
    await connectProtocol(clientEndpoint, clientProtocol)
    await serverFactory.clientDisconnected
    await serverPort.stopListening()



from slots_attributes_1 import ExampleElement

from twisted.web.template import flattenString


def renderDone(output):
    print(output)


flattenString(None, ExampleElement()).addCallback(renderDone)



from MulticastCommon import group, interface, port

from twisted.internet import reactor
from twisted.internet.protocol import DatagramProtocol


class MulticastPingClient(DatagramProtocol):
    def startProtocol(self):
        # Join the multicast address, so we can receive replies:
        self.transport.joinGroup(group)
        # Send to our group:port pair - all listeners on the multicast address
        # (including us) will receive this message.
        self.transport.write(b"Client: Ping", (group, port))

    def datagramReceived(self, datagram, address):
        print(f"Datagram {repr(datagram)} received from {repr(address)}")


reactor.listenMulticast(
    port, MulticastPingClient(), listenMultiple=True, interface=interface
)
reactor.run()



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


"""Server for PB benchmark."""

from zope.interface import implementer

from twisted.cred.portal import IRealm
from twisted.internet import reactor
from twisted.spread import pb


class PBBenchPerspective(pb.Avatar):
    callsPerSec = 0

    def __init__(self):
        pass

    def perspective_simple(self):
        self.callsPerSec = self.callsPerSec + 1
        return None

    def printCallsPerSec(self):
        print("(s) cps:", self.callsPerSec)
        self.callsPerSec = 0
        reactor.callLater(1, self.printCallsPerSec)

    def perspective_complexTypes(self):
        return ["a", 1, 1, 1.0, [], ()]


@implementer(IRealm)
class SimpleRealm:
    def requestAvatar(self, avatarId, mind, *interfaces):
        if pb.IPerspective in interfaces:
            p = PBBenchPerspective()
            p.printCallsPerSec()
            return pb.IPerspective, p, lambda: None
        else:
            raise NotImplementedError("no interface")


def main():
    from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
    from twisted.cred.portal import Portal

    portal = Portal(SimpleRealm())
    checker = InMemoryUsernamePasswordDatabaseDontUse()
    checker.addUser(b"benchmark", b"benchmark")
    portal.registerChecker(checker)
    reactor.listenTCP(8787, pb.PBServerFactory(portal))
    reactor.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

import echoclient

from twisted.internet import defer, endpoints, protocol, ssl, task
from twisted.python.modules import getModule


async def main(reactor):
    factory = protocol.Factory.forProtocol(echoclient.EchoClient)
    certData = getModule(__name__).filePath.sibling("public.pem").getContent()
    authData = getModule(__name__).filePath.sibling("server.pem").getContent()
    clientCertificate = ssl.PrivateCertificate.loadPEM(authData)
    authority = ssl.Certificate.loadPEM(certData)
    options = ssl.optionsForClientTLS(
        "example.com",
        trustRoot=authority,
        clientCertificate=clientCertificate,
    )
    tcpEndpoint = endpoints.HostnameEndpoint(reactor, "localhost", 8000)
    endpoint = endpoints.wrapClientTLS(options, tcpEndpoint, reactor)
    done = defer.Deferred()
    echoClient = await endpoint.connect(factory)
    echoClient.dataReceived = lambda data: done.callback(data)
    print(await done)


if __name__ == "__main__":
    import ssl_clientauth_client

    task.react(ssl_clientauth_client.main)



import io

from ad_hoc import AdHoc

from twisted.logger import globalLogPublisher, jsonFileLogObserver

globalLogPublisher.addObserver(jsonFileLogObserver(open("log.json", "a")))

AdHoc(3, 4).logMessage()



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

import pytest

from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.trial.reporter import TestResult
from twisted.trial.unittest import TestCase


class BenchmarkMixin:
    def testPlain(self):
        pass

    def testResolvedDeferred(self):
        return succeed(None)

    def testUnresolvedDeferred(self):
        d = Deferred()
        reactor.callLater(0, d.callback, None)
        return d

    def testAddCleanupPlain(self):
        self.addCleanup(lambda: None)

    def testAddCleanupResolvedDeferred(self):
        self.addCleanup(lambda: succeed(None))

    def testAddCleanupUnresolvedDeferred(self):
        def cleanup():
            d = Deferred()
            reactor.callLater(0, d.callback, None)
            return d

        self.addCleanup(cleanup)


class BenchmarkTestsNoSetUp(BenchmarkMixin, TestCase):
    pass


class BenchmarkTestsSetUpTearDownPlain(BenchmarkMixin, TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


class BenchmarkTestsSetUpTearDownResolvedDeferred(BenchmarkMixin, TestCase):
    def setUp(self):
        return succeed(None)

    def tearDown(self):
        return succeed(None)


class BenchmarkTestsSetUpTearDownUnresolvedDeferred(BenchmarkMixin, TestCase):
    def setUp(self):
        d = Deferred()
        reactor.callLater(0, d.callback, None)
        return d

    def tearDown(self):
        d = Deferred()
        reactor.callLater(0, d.callback, None)
        return d


@pytest.mark.parametrize(
    "method_name",
    [
        "testPlain",
        "testResolvedDeferred",
        "testUnresolvedDeferred",
        "testAddCleanupPlain",
        "testAddCleanupResolvedDeferred",
        "testAddCleanupUnresolvedDeferred",
    ],
)
def test_main_method(benchmark, method_name):
    """Measure the speed of tests with different implementations of test method"""

    def go():
        result = TestResult()
        BenchmarkTestsNoSetUp(method_name).run(result)

    benchmark(go)


@pytest.mark.parametrize(
    "test_class",
    [
        BenchmarkTestsSetUpTearDownPlain,
        BenchmarkTestsSetUpTearDownResolvedDeferred,
        BenchmarkTestsSetUpTearDownUnresolvedDeferred,
    ],
)
def test_setup_teardown(benchmark, test_class):
    """Measure the speed of tests with different implementations of setUp and tearDown"""

    def go():
        result = TestResult()
        test_class("testPlain").run(result)

    benchmark(go)
